﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=8.9
@EndOfDesignText@
' Converts an .ics file's VTODO (task) entries to VEVENT (calendar) entries.

Sub Class_Globals

	Private fx As JFX

	Private inputList As List
	Private originalDateFormat As String = DateTime.DateFormat
	Private originalTimeFormat As String = DateTime.TimeFormat
	Private initError As String = ""

End Sub

' Initializes the object. You can add parameters to this method if needed.
Public Sub Initialize(inputDir As String, inputFile As String)

	If File.Exists(inputDir, inputFile) = False Then
		initError = "The input file does not exist"
		Log("wmICSvtodoToVevent - Initialize: input file does not exist")
		Return
	End If

	inputList = File.ReadList(inputDir, inputFile)	

End Sub

' Returns "" if initialization completed successfully and an error message if it failed
Public Sub InitializedOK As String

	Return initError

End Sub

' - outputDir: the output file's directory
' - outputFile: the output file name
' - removeEvents: if True, the VEVENT (calendar) entries will be removed (at least one of removeEvents and removeTodos must be False)
' - removeTodos: if True, the VTODO entries will be removed instead of transformed to VEVENT entries (at least one of removeEvents and removeTodos must be False)
' - prefixForTitle: will be prepended to the 'SUMMARY' entries (which contain the event/task's title)
'   so that it's easy to distinguish tasks in the calendar view on the Android device, e.g. use "(T)"
' - TZID: see https://icalendar.org/iCalendar-RFC-5545/3-2-19-time-zone-identifier.html ;
'   if "", "Europe/Paris" is used
' - minutes: the number of minutes to assign to the calendar entry that comes from a task that
'   has a start time; if less than 5, 5 is used; tasks without a start time will be converted
'   into all-day calendar events
' - returns "" if something went wrong, or the latest 'LAST-MODIFIED' value, e.g. "20040302T225559Z"
Public Sub VTODOtoVEVENT(outputDir As String, outputFile As String, _
						removeEvents As Boolean, removeTodos As Boolean, _
						prefixForTitle As String, TZID As String, minutes As Int) As String

	Dim i As Int
	Dim vtodoStarted As Boolean = False
	Dim veventStarted As Boolean = False
	Dim seenDTstart As Boolean = False
	Dim outputList As List
	Dim outputListEntryLastModifiedDate As Int = -1
	Dim forcedLastModified As String = "LAST-MODIFIED:" & GetDTend(DateTime.Now) & "Z"
	Dim latestLastModified As String = ""

	If initError <> "" Then
		Log("wmICSvtodoToVevent: initialize appears to have failed")
		Return ""
	End If

	If removeEvents And removeTodos Then
		Log("wmICSvtodoToVevent: neither events nor todos would be processed")
		Return ""
	End If

	If minutes < 5 Then minutes = 5
	If TZID = "" Then TZID = "Europe/Paris"

	' Process the data
	outputList.Initialize
	For i = 0 To (inputList.size - 1)
		Dim line As String = inputList.Get(i)
		' Save latest 'LAST-MODIFIED'?
		If line.StartsWith("LAST-MODIFIED:") Then
			If line.CompareTo(latestLastModified) > 0 Then latestLastModified = line
		End If
		If removeEvents And line.StartsWith("BEGIN:VEVENT") Then ' Events are ignored if 'removeEvents' is True
			veventStarted = True
		Else If veventStarted And line.StartsWith("END:VEVENT") Then ' End of the Event
			veventStarted = False
		Else If line.StartsWith("BEGIN:VTODO") Then ' A task is found
			vtodoStarted = True
			If removeTodos = False Then
				outputList.Add(line.Replace("VTODO", "VEVENT"))
				outputList.Add("TRANSP:TRANSPARENT") ' Didn't see this in my VTODO data, only in my VEVENT data, so adding it artificially here
				outputList.Add("CLASS:PRIVATE") ' Didn't see this in my VTODO data, only in my VEVENT data, so adding it artificially here
			End If
		Else If line.StartsWith("END:VTODO") Then ' Complete the task's data
			If removeTodos = False Then
				If seenDTstart = False Then ' Event entries must have a DTSTART
					outputList.Add("DTSTART;VALUE=DATE:" & TicksToYYYYMMDD(DateTime.Now)) ' We're using 'today' as default
					outputList.Add("DTEND;VALUE=DATE:" & TicksToYYYYMMDD(DateTime.Add(DateTime.Now, 0, 0, 1))) ' We're making this an all-day event
					' Ensure the entry will get updated on the device by forcing 'LAST-MODIFIED'
					If outputListEntryLastModifiedDate < 0 Then
						outputList.Add(forcedLastModified)
					Else
						outputList.InsertAt(outputListEntryLastModifiedDate, forcedLastModified)
						outputList.RemoveAt(outputListEntryLastModifiedDate + 1)
					End If
				End If
				outputList.Add(line.Replace("VTODO", "VEVENT"))
			End If
			vtodoStarted = False
			seenDTstart = False
			outputListEntryLastModifiedDate = -1
		Else If vtodoStarted Then
			If removeTodos = False Then
				If line.CompareTo("SUMMARY:") = 0 Then
					outputList.Add("SUMMARY:" & prefixForTitle) ' The Summary (title) is empty; just use the prefix to distinguish tasks vs. appointments
				Else If line.StartsWith("SUMMARY:") Then
					outputList.Add("SUMMARY:" & prefixForTitle & line.SubString(8)) ' Use the prefix to distinguish tasks vs. appointments
				Else If line.StartsWith("LAST-MODIFIED:") Then
					outputList.Add(line)
					outputListEntryLastModifiedDate = outputList.Size - 1 ' Save the entry's position in the list; we may need it when processing "END:VTODO"
				Else If line.StartsWith("DTSTART") Then
					seenDTstart = True
					HandleDTstart(outputList, line, TZID, minutes) ' Calculate and add the DTEND entry
				Else
					outputList.Add(line) ' Just copy this line
				End If
			End If
		Else If veventStarted = False Then ' If it were True, it means that we're ignoring Events
			outputList.Add(line)
		End If
	Next

	' Write the output file
	If File.Exists(outputDir, outputFile) Then File.Delete(outputDir, outputFile)
	File.WriteList(outputDir, outputFile, outputList)

	' Done
	If latestLastModified <> "" Then latestLastModified = latestLastModified.SubString(14) ' Remove "LAST-MODIFIED:" from the value to return
	Return latestLastModified

End Sub

' Adds DTSTART and DTEND lines based on the DTSTART value
Private Sub HandleDTstart(outputList As List, line As String, TZID As String, minutes As Int)

	' Values I observed in my data for an EVENT (calendar) on 5 September 2020, all day:
	' - DTSTART;VALUE=DATE:20200905
	' - DTEND;VALUE=DATE:20200906

	' Values I observed in my data for an EVENT (calendar) on 5 September 2020 from 17:00 to 23:00:
	' - DTSTART;TZID=Europe/Paris:20200905T170000
	' - DTEND;TZID=Europe/Paris:20200905T230000

	' Values I observed in my data for a TODO (task) on 5 September 2020:
	' - DTSTART;TZID=Europe/Paris:20200905T000000 <<< 'without' start time, i.e. start time set to "00:00"
	' - DTSTART;TZID=Europe/Paris:20200905T154500 <<< with start time set to "15:45"

	Dim idxColon As Int = line.LastIndexOf(":")
	Dim sDateTime As String = line.SubString(idxColon + 1) ' Still includes the 'T', which will be removed later
	Dim idxT As Int = sDateTime.IndexOf("T")
	Dim sDateStart As String = sDateTime.SubString2(0, idxT)
	Dim lDateStart As Long = YYYYMMDDtoTicks(sDateStart)
	Dim sTimeStart As String = sDateTime.SubString(idxT + 1)
	sDateTime = sDateTime.Replace("T", "")

	If sDateTime.EndsWith("000000") Then
		' Task without start time: create an all-day calendar entry. Timezone is irrelevant for all-day entries.
		outputList.Add("DTSTART;VALUE=DATE:" & sDateStart)
		outputList.Add("DTEND;VALUE=DATE:" & TicksToYYYYMMDD(DateTime.Add(lDateStart, 0, 0, 1)))
	Else
		' Task with a start time: create a calendar entry for the specified Timezone with duration='minutes'.
		outputList.Add("DTSTART;TZID=" & TZID & ":" & sDateStart & "T" & sTimeStart)
		Dim p As Period
		p.Initialize
		p.Minutes = minutes
		Dim lDateTimeEnd As Long = DateUtils.AddPeriod(DTstartToTicks(sDateStart, sTimeStart), p)
		outputList.Add("DTEND;TZID=" & TZID & ":" & GetDTend(lDateTimeEnd))
	End If

End Sub

Private Sub YYYYMMDDtoTicks(val As String) As Long

	DateTime.DateFormat = "yyyyMMdd"
	Dim dateOut As Long = DateTime.DateParse(val)
	DateTime.DateFormat = originalDateFormat
	Return dateOut

End Sub

Private Sub TicksToYYYYMMDD(val As Long) As String

	DateTime.DateFormat = "yyyyMMdd"
	Dim YYYYMMDDout As String = DateTime.Date(val)
	DateTime.DateFormat = originalDateFormat
	Return YYYYMMDDout

End Sub

Private Sub DTstartToTicks(Dval As String, Tval As String) As Long

	DateTime.DateFormat = "yyyyMMdd"
	DateTime.TimeFormat = "HHmmss"
	Dim dateTimeOut As Long = DateTime.DateTimeParse(Dval, Tval)
	DateTime.DateFormat = originalDateFormat
	DateTime.TimeFormat = originalTimeFormat
	Return dateTimeOut

End Sub

Private Sub GetDTend(val As Long) As String

	DateTime.DateFormat = "yyyyMMdd"
	DateTime.TimeFormat = "HHmmss"
	Dim strOut As String = DateTime.Date(val) & "T" & DateTime.Time(val)
	DateTime.DateFormat = originalDateFormat
	DateTime.TimeFormat = originalTimeFormat
	Return strOut

End Sub